/*************************************************** 
  www.buydisplay.com
 ****************************************************/

#include "TFTM2.08-1.h"
#include <avr/pgmspace.h>
#include <limits.h>
#include "pins_arduino.h"
#include "wiring_private.h"
#include <SPI.h>

// Constructor when using software SPI.  All output pins are configurable.
NV3203B::NV3203B(int8_t cs, int8_t dc, int8_t mosi,
				   int8_t sclk, int8_t rst) : ERGFX(NV3203B_TFTWIDTH, NV3203B_TFTHEIGHT) {
  _cs   = cs;
  _dc   = dc;
  _mosi  = mosi;
  _sclk = sclk;
  _rst  = rst;
  hwSPI = false;
}


// Constructor when using hardware SPI.  Faster, but must use SPI pins
// specific to each board type (e.g. 11,13 for Uno, 51,52 for Mega, etc.)
NV3203B::NV3203B(int8_t cs, int8_t dc, int8_t rst) : ERGFX(NV3203B_TFTWIDTH, NV3203B_TFTHEIGHT) {
  _cs   = cs;
  _dc   = dc;
  _rst  = rst;
  hwSPI = true;
  _mosi  = _sclk = 0;
}

void NV3203B::spiwrite(uint8_t c) {

  //Serial.print("0x"); Serial.print(c, HEX); Serial.print(", ");

  if (hwSPI) {
#if defined (__AVR__)
      uint8_t backupSPCR = SPCR;
    SPCR = mySPCR;
    SPDR = c;
    while(!(SPSR & _BV(SPIF)));
    SPCR = backupSPCR;
#elif defined(TEENSYDUINO)
    SPI.transfer(c);
#elif defined (__arm__)
    SPI.setClockDivider(11); // 8-ish MHz (full! speed!)
    SPI.setBitOrder(MSBFIRST);
    SPI.setDataMode(SPI_MODE0);
    SPI.transfer(c);
#endif
  } else {
    // Fast SPI bitbang swiped from LPD8806 library
    for(uint8_t bit = 0x80; bit; bit >>= 1) {
      if(c & bit) {
	//digitalWrite(_mosi, HIGH); 
	*mosiport |=  mosipinmask;
      } else {
	//digitalWrite(_mosi, LOW); 
	*mosiport &= ~mosipinmask;
      }
      //digitalWrite(_sclk, LOW);
      *clkport &= ~clkpinmask;
       //digitalWrite(_sclk, HIGH);
      *clkport |=  clkpinmask;
    }
  }
}


void NV3203B::writecommand(uint8_t c) {
  *dcport &=  ~dcpinmask;
  //digitalWrite(_dc, LOW);
  //*clkport &= ~clkpinmask; // clkport is a NULL pointer when hwSPI==true
  //digitalWrite(_sclk, LOW);
  *csport &= ~cspinmask;
  //digitalWrite(_cs, LOW);

  spiwrite(c);

  *csport |= cspinmask;
  //digitalWrite(_cs, HIGH);
}


void NV3203B::writedata(uint8_t c) {
  *dcport |=  dcpinmask;
  //digitalWrite(_dc, HIGH);
  //*clkport &= ~clkpinmask; // clkport is a NULL pointer when hwSPI==true
  //digitalWrite(_sclk, LOW);
  *csport &= ~cspinmask;
  //digitalWrite(_cs, LOW);
  
  spiwrite(c);

  //digitalWrite(_cs, HIGH);
  *csport |= cspinmask;
} 

// If the SPI library has transaction support, these functions
// establish settings and protect from interference from other
// libraries.  Otherwise, they simply do nothing.
#ifdef SPI_HAS_TRANSACTION
static inline void spi_begin(void) __attribute__((always_inline));
static inline void spi_begin(void) {
  SPI.beginTransaction(SPISettings(8000000, MSBFIRST, SPI_MODE0));
}
static inline void spi_end(void) __attribute__((always_inline));
static inline void spi_end(void) {
  SPI.endTransaction();
}
#else
#define spi_begin()
#define spi_end()
#endif


void NV3203B::begin(void) {
  if (_rst > 0) {
    pinMode(_rst, OUTPUT);
    digitalWrite(_rst, LOW);
  }

  pinMode(_dc, OUTPUT);
  pinMode(_cs, OUTPUT);
  
  csport    = portOutputRegister(digitalPinToPort(_cs));
  cspinmask = digitalPinToBitMask(_cs);
  dcport    = portOutputRegister(digitalPinToPort(_dc));
  dcpinmask = digitalPinToBitMask(_dc);

  if(hwSPI) { // Using hardware SPI
#if defined (__AVR__)
    SPI.begin();
    SPI.setClockDivider(SPI_CLOCK_DIV2); // 8 MHz (full! speed!)
    SPI.setBitOrder(MSBFIRST);
    SPI.setDataMode(SPI_MODE0);
    mySPCR = SPCR;
#elif defined(TEENSYDUINO)
    SPI.begin();
    SPI.setClockDivider(SPI_CLOCK_DIV2); // 8 MHz (full! speed!)
    SPI.setBitOrder(MSBFIRST);
    SPI.setDataMode(SPI_MODE0);
#elif defined (__arm__)
      SPI.begin();
      SPI.setClockDivider(11); // 8-ish MHz (full! speed!)
      SPI.setBitOrder(MSBFIRST);
      SPI.setDataMode(SPI_MODE0);
#endif
  } else {
    pinMode(_sclk, OUTPUT);
    pinMode(_mosi, OUTPUT);

    clkport     = portOutputRegister(digitalPinToPort(_sclk));
    clkpinmask  = digitalPinToBitMask(_sclk);
    mosiport    = portOutputRegister(digitalPinToPort(_mosi));
    mosipinmask = digitalPinToBitMask(_mosi);
    *clkport   &= ~clkpinmask;
    *mosiport  &= ~mosipinmask;
  }

  // toggle RST low to reset
  if (_rst > 0) {
    digitalWrite(_rst, HIGH);
    delay(5);
    digitalWrite(_rst, LOW);
    delay(20);
    digitalWrite(_rst, HIGH);
    delay(150);
  }

  if (hwSPI) spi_begin();

 //----------------Star Initial Sequence-------//
 	writecommand(0xff);
	writedata(0xa5); //
 	writecommand(0x3E);
	writedata(0x08); //
 	writecommand(0x3A);
	writedata(0x55);
 	writecommand(0x82);
	writedata(0x00);
 	writecommand(0x98);
	writedata(0x00);
 	writecommand(0x63);
	writedata(0x10);
 	writecommand(0x64);
	writedata(0x10);
 	writecommand(0xB4);
	writedata(0x34); // 34 55
 	writecommand(0xB5);
	writedata(0x30);
 	writecommand(0x83);
	writedata(0x03);
 	writecommand(0x86); //
	writedata(0x00);
 	writecommand(0x87);
	writedata(0x16);
 	writecommand(0x88); // VCOM
	writedata(0x09);
 	writecommand(0x89); //
	writedata(0x2f);
 	writecommand(0x93); //
	writedata(0x63);
 	writecommand(0x96);
	writedata(0x81);
 	writecommand(0xC3);
	writedata(0x01); //
 	writecommand(0xE6);
	writedata(0x00);
 	writecommand(0x99);
	writedata(0x01);
    ////////////////////////gamma_set//////////////////////////////////////
 	writecommand(0x70);
	writedata(0x07); // VRP 0
 	writecommand(0x71);
	writedata(0x1b); // VRP 1
 	writecommand(0x72);
	writedata(0x08); // VRP 2
 	writecommand(0x73);
	writedata(0x0f); // VRP 3
 	writecommand(0x74);
	writedata(0x16); // VRP 6
 	writecommand(0x75);
	writedata(0x1A); // VRP 8
 	writecommand(0x76);
	writedata(0x3c); // VRP 10
 	writecommand(0x77);
	writedata(0x0a); // VRP 14
 	writecommand(0x78);
	writedata(0x05); // VRP 17
 	writecommand(0x79);
	writedata(0x3a); // VRP 21
 	writecommand(0x7a);
	writedata(0x06); // VRP 23
 	writecommand(0x7b);
	writedata(0x0b); // VRP 25
 	writecommand(0x7c);
	writedata(0x12); // VRP 28
 	writecommand(0x7d);
	writedata(0x0b); // VRP 29
 	writecommand(0x7e);
	writedata(0x0a); // VRP 30
 	writecommand(0x7f);
	writedata(0x08); // VRP 31
 	writecommand(0xa0);
	writedata(0x0b); // VRN 0
 	writecommand(0xa1);
	writedata(0x36); // VRN 1
 	writecommand(0xa2);
	writedata(0x0b); // VRN 2
 	writecommand(0xa3);
	writedata(0x0c); // VRN 3
 	writecommand(0xa4);
	writedata(0x08); // VRN 6
 	writecommand(0xa5);
	writedata(0x22); // VRN 8
 	writecommand(0xa6);
	writedata(0x43); // VRN 10
 	writecommand(0xa7);
	writedata(0x04); // VRN 14
 	writecommand(0xa8);
	writedata(0x05); // VRN 17
 	writecommand(0xa9);
	writedata(0x3F); // VRN 21
 	writecommand(0xaa);
	writedata(0x0A); // VRN 23
 	writecommand(0xab);
	writedata(0x11); // VRN 25
 	writecommand(0xac);
	writedata(0x0d); // VRN 28
 	writecommand(0xad);
	writedata(0x06); // VRN 29
 	writecommand(0xae);
	writedata(0x3b); // VRN 30
 	writecommand(0xaf);
	writedata(0x07); // VRN 31
    //////////////////////////////////////////////////////////////////
 	writecommand(0xff);
	writedata(0x00);

 	writecommand(0x36); // MX, MY, RGB mode
   	writedata(0x48);

        writecommand(0x11);    //Exit Sleep 
        if (hwSPI) spi_end();
        delay(120); 		
        if (hwSPI) spi_begin();
        writecommand(0x29);    //Display on 
        if (hwSPI) spi_end();

}


void NV3203B::setAddrWindow(uint16_t x0, uint16_t y0, uint16_t x1,
 uint16_t y1) {

  writecommand(NV3203B_CASET); // Column addr set
  writedata(x0 >> 8);
  writedata(x0);     // XSTART 
  writedata(x1 >> 8);
  writedata(x1);     // XEND

  writecommand(NV3203B_PASET); // Row addr set
  writedata(y0>>8);
  writedata(y0);     // YSTART
  writedata(y1>>8);
  writedata(y1);     // YEND

  writecommand(NV3203B_RAMWR); // write to RAM
}


void NV3203B::pushColor(uint16_t color) {
  if (hwSPI) spi_begin();
  //digitalWrite(_dc, HIGH);
  *dcport |=  dcpinmask;
  //digitalWrite(_cs, LOW);
  *csport &= ~cspinmask;

  spiwrite(color >> 8);
  spiwrite(color);

  *csport |= cspinmask;
  //digitalWrite(_cs, HIGH);
  if (hwSPI) spi_end();
}

void NV3203B::drawPixel(int16_t x, int16_t y, uint16_t color) {

  if((x < 0) ||(x >= _width) || (y < 0) || (y >= _height)) return;

  if (hwSPI) spi_begin();
  setAddrWindow(x,y,x+1,y+1);

  //digitalWrite(_dc, HIGH);
  *dcport |=  dcpinmask;
  //digitalWrite(_cs, LOW);
  *csport &= ~cspinmask;

  spiwrite(color >> 8);
  spiwrite(color);

  *csport |= cspinmask;
  //digitalWrite(_cs, HIGH);
  if (hwSPI) spi_end();
}


void NV3203B::drawFastVLine(int16_t x, int16_t y, int16_t h,
 uint16_t color) {

  // Rudimentary clipping
  if((x >= _width) || (y >= _height)) return;

  if((y+h-1) >= _height) 
    h = _height-y;

  if (hwSPI) spi_begin();
  setAddrWindow(x, y, x, y+h-1);

  uint8_t hi = color >> 8, lo = color;

  *dcport |=  dcpinmask;
  //digitalWrite(_dc, HIGH);
  *csport &= ~cspinmask;
  //digitalWrite(_cs, LOW);

  while (h--) {
    spiwrite(hi);
    spiwrite(lo);
  }
  *csport |= cspinmask;
  //digitalWrite(_cs, HIGH);
  if (hwSPI) spi_end();
}


void NV3203B::drawFastHLine(int16_t x, int16_t y, int16_t w,
  uint16_t color) {

  // Rudimentary clipping
  if((x >= _width) || (y >= _height)) return;
  if((x+w-1) >= _width)  w = _width-x;
  if (hwSPI) spi_begin();
  setAddrWindow(x, y, x+w-1, y);

  uint8_t hi = color >> 8, lo = color;
  *dcport |=  dcpinmask;
  *csport &= ~cspinmask;
  //digitalWrite(_dc, HIGH);
  //digitalWrite(_cs, LOW);
  while (w--) {
    spiwrite(hi);
    spiwrite(lo);
  }
  *csport |= cspinmask;
  //digitalWrite(_cs, HIGH);
  if (hwSPI) spi_end();
}

void NV3203B::fillScreen(uint16_t color) {
  fillRect(0, 0,  _width, _height, color);
}

// fill a rectangle
void NV3203B::fillRect(int16_t x, int16_t y, int16_t w, int16_t h,
  uint16_t color) {

  // rudimentary clipping (drawChar w/big text requires this)
  if((x >= _width) || (y >= _height)) return;
  if((x + w - 1) >= _width)  w = _width  - x;
  if((y + h - 1) >= _height) h = _height - y;

  if (hwSPI) spi_begin();
  setAddrWindow(x, y, x+w-1, y+h-1);

  uint8_t hi = color >> 8, lo = color;

  *dcport |=  dcpinmask;
  //digitalWrite(_dc, HIGH);
  *csport &= ~cspinmask;
  //digitalWrite(_cs, LOW);

  for(y=h; y>0; y--) {
    for(x=w; x>0; x--) {
      spiwrite(hi);
      spiwrite(lo);
    }
  }
  //digitalWrite(_cs, HIGH);
  *csport |= cspinmask;
  if (hwSPI) spi_end();
}


// Pass 8-bit (each) R,G,B, get back 16-bit packed color
uint16_t NV3203B::color565(uint8_t r, uint8_t g, uint8_t b) {
  return ((r & 0xF8) << 8) | ((g & 0xFC) << 3) | (b >> 3);
}

 

